#include <ros/ros.h>
#include <tf/tf.h>
#include <tf/transform_listener.h>
#include <geometry_msgs/Twist.h>
#include <geometry_msgs/PoseStamped.h>
#include <geometry_msgs/PoseArray.h>
#include <geometry_msgs/PoseWithCovarianceStamped.h>
#include <move_base_msgs/MoveBaseAction.h>

#include <cmath>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include <thread>

typedef struct {
    double x;
    double y;
    double yaw;
} WayPoint;


bool g_forward_wp_flag = false;
std::vector<WayPoint> g_wp_array;
std::vector<WayPoint>::iterator g_wp_array_iter;

//別スレッドで監視
//Key入力(enter keyを想定)がある度に、waypointを次に進める
void ForwardWaypointByKeyboardInterrupt() {
    while (1) {
        getchar();
        g_forward_wp_flag = true;
        std::cout << "KeyboardInterrupt" << std::endl;
    }
}

std::vector<WayPoint>::iterator SearchNearWaypointIterator(WayPoint target_wp)
{
    const auto most_near_iter = std::min_element(std::begin(g_wp_array), std::end(g_wp_array), [&target_wp](WayPoint lhs, WayPoint rhs) {
        return std::hypot(target_wp.x - lhs.x, target_wp.y - lhs.y) < std::hypot(target_wp.x - rhs.x, target_wp.y - rhs.y);
    });
    return most_near_iter;
}

//rvizの2D Pose Estimateを設定された場合、次のwaypointをもっとも近い位置に変更する
//最も近いwaypointの探索は、initial_poseのx,yのみに依存する
void InitialPoseCallBack(const geometry_msgs::PoseWithCovarianceStamped::ConstPtr& initial_pose)
{
    std::cout << "Initial pose" << std::endl;
    g_forward_wp_flag = true;

    WayPoint tmp_wp;
    tmp_wp.x = initial_pose->pose.pose.position.x;
    tmp_wp.y = initial_pose->pose.pose.position.y;

    g_wp_array_iter = SearchNearWaypointIterator(tmp_wp);
}

std::vector<WayPoint> ReadWayPointFile(const std::string a_file_name) {
    std::vector<WayPoint> g_wp_array;
    std::ifstream wp_file(a_file_name);
    if (!wp_file) {
        std::cerr << "ERROR: Can't open waypoint file!!" << std::endl;
        return g_wp_array;
    }

    std::string reading_line;
    while (1) {
        std::getline(wp_file, reading_line);
        if (reading_line.empty()) {
            std::cout << "End of file." << std::endl;
            break;
        }
        std::stringstream line_ss(reading_line);

        WayPoint tmp_wp;
        line_ss >> tmp_wp.x >> tmp_wp.y >> tmp_wp.yaw;

        g_wp_array.push_back(tmp_wp);
    }

    return g_wp_array;
}

geometry_msgs::PoseArray ConvertToWayPointMsg(const std::vector<WayPoint> g_wp_array) {
    geometry_msgs::PoseArray g_wp_array_msg;
    g_wp_array_msg.header.stamp = ros::Time::now();;
    g_wp_array_msg.header.frame_id = "map";
    
    for(auto &wp : g_wp_array) {
        geometry_msgs::Pose wp_msg;
        wp_msg.position.x = wp.x;
        wp_msg.position.y = wp.y;
        wp_msg.position.z = 0.0;
        wp_msg.orientation = tf::createQuaternionMsgFromYaw(wp.yaw);        
        g_wp_array_msg.poses.push_back(wp_msg);
    }
    
    return g_wp_array_msg;
}

int main(int argc, char **argv) {

    // 固定小数点記法に設定
    std::cout << std::fixed;
    std::cerr << std::fixed;

    // 引数が正しくない場合は使い方を表示
    if (argc < 2) {
        std::cout << "USAGE:" << std::endl;
        std::cout << "  $ " << argv[0] << " <wp_file_path>" << std::endl;
        std::cout << "  e.g.)  $ " << argv[0] << " wp.txt" << std::endl;
        return -1;
    }


    ros::init(argc, argv, "waypoint_publisher");
    ros::NodeHandle public_node_handle;
    ros::NodeHandle private_node_handle("~");

    ros::Publisher cmd_publisher = public_node_handle.advertise<geometry_msgs::Twist>("cmd_vel", 1);
    ros::Publisher nav_goal_publisher = public_node_handle.advertise<geometry_msgs::PoseStamped>("move_base_simple/goal", 1);
    ros::Publisher wp_publisher = public_node_handle.advertise<geometry_msgs::PoseArray>("waypoints", 1);

    ros::Subscriber initial_pose_subscriber = public_node_handle.subscribe<geometry_msgs::PoseWithCovarianceStamped>("initialpose", 1, InitialPoseCallBack);

    double goal_tolerance_m = 2.0;
    double skip_time_sec = 15.0;
    private_node_handle.param("g_tolerance_m", goal_tolerance_m, goal_tolerance_m);
    private_node_handle.param("skip_time_sec", skip_time_sec, skip_time_sec);

    std::thread keyboad_thread(ForwardWaypointByKeyboardInterrupt);
    keyboad_thread.detach();

    g_wp_array = ReadWayPointFile(argv[1]);
    g_wp_array_iter = std::begin(g_wp_array);
    const geometry_msgs::PoseArray g_wp_array_msg = ConvertToWayPointMsg(g_wp_array);

    tf::TransformListener tf_listener;
    ros::Duration(5).sleep();

    while (ros::ok() && g_wp_array_iter != std::end(g_wp_array)) {
        static double wp_x_m = 0, wp_y_m = 0, wp_yaw_rad = 0;
        static ros::Time start_time = ros::Time::now(); //waypointが変わるたびに更新される

        //一定時間waypointに辿りつけなければ、次へ進む
        if (!g_forward_wp_flag && (ros::Time::now() - start_time) > ros::Duration(skip_time_sec)) {
            std::cout << "Time out" << std::endl;
            g_forward_wp_flag = true;
        }

        //waypointに辿りつけば、次へ進む
        if (!g_forward_wp_flag) {
            tf::StampedTransform transform;
            try {
                tf_listener.waitForTransform("map", "base_link", ros::Time(0), ros::Duration(1.0));
                tf_listener.lookupTransform("map", "base_link", ros::Time(0), transform);
            }
            catch (tf::TransformException ex) {
                ROS_ERROR("%s", ex.what());
                continue;
            }

            const double diff_x_m = wp_x_m - transform.getOrigin().x();
            const double diff_y_m = wp_y_m - transform.getOrigin().y();
            const double diff_dis_m = std::hypot(diff_x_m, diff_y_m);

            if (diff_dis_m < goal_tolerance_m) {
                std::cout << "Reach waypoint" << std::endl;
                g_forward_wp_flag = true;
            }
        }

        //次のwaypointをpublishする
        if (g_forward_wp_flag) {
            start_time = ros::Time::now();

            ++g_wp_array_iter;
            wp_x_m = g_wp_array_iter->x;
            wp_y_m = g_wp_array_iter->y;

            geometry_msgs::PoseStamped current_goal;
            current_goal.header.stamp = ros::Time::now();;
            current_goal.header.frame_id = "map";
            current_goal.pose.position.x = wp_x_m;
            current_goal.pose.position.y = wp_y_m;
            current_goal.pose.position.z = 0.0;
            current_goal.pose.orientation = tf::createQuaternionMsgFromYaw(g_wp_array_iter->yaw);

            nav_goal_publisher.publish(current_goal);

            g_forward_wp_flag = false;
        }

        wp_publisher.publish(g_wp_array_msg);
        ros::spinOnce();
        ros::Duration(0.1).sleep();
    }

    // ゴールした後で停止するためにコマンド発行
    geometry_msgs::Twist cmd_vel;
    cmd_vel.linear.x = 0.0;
    cmd_vel.linear.y = 0.0;
    cmd_vel.linear.z = 0.0;
    cmd_vel.angular.x = 0.0;
    cmd_vel.angular.y = 0.0;
    cmd_vel.angular.z = 0.0;
    cmd_publisher.publish(cmd_vel);
    ros::spinOnce();

    return 0;
}
